(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Efficient_RPO
imports
  RPO
  RPO_Impl
  Multiset_Extension_Impl
  List2
  SubList
  Memories
begin

(** A memoized computation of type 'a is a computation that,
    given an environment of type 'm computes and 'a and a
    new environment of type 'm. **)

type_synonym ('m, 'a) memo = "'m \<Rightarrow> ('a \<times> 'm)"

(** Type synonyms for precedence and tagging functions. **)
type_synonym 'f prec_fct = "('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool \<times> bool) \<times> ('f \<times> nat \<Rightarrow> bool)"
type_synonym 'f status_fct = "'f \<times> nat \<Rightarrow> order_tag"

(** Efficient versions of all and any for the specific cases that
    we are interested in. **)

fun any_nstri_efficient_m :: "('a \<Rightarrow> ('m, bool \<times> bool) memo) \<Rightarrow> 'a list \<Rightarrow> ('m, bool) memo"
where
  "any_nstri_efficient_m f []       m = (False, m)" |
  "any_nstri_efficient_m f (a # as) m = (let ((_, ns), m') = f a m in
     if ns then (True, m') else any_nstri_efficient_m f as m')"

fun all_stri_efficient_m :: "('a \<Rightarrow> ('m, bool \<times> bool) memo) \<Rightarrow> 'a list \<Rightarrow> ('m, bool) memo"
where
  "all_stri_efficient_m f []       m = (True, m)" |
  "all_stri_efficient_m f (a # as) m = (let ((s, _), m') = f a m in
     if s then all_stri_efficient_m f as m' else (False, m'))"

(** congruence lemmas needed for rpo's termination proof **)

lemma any_nstri_efficient_m[fundef_cong]:
  assumes "\<And> s m. s \<in> set ss \<Longrightarrow> f s m = g s m"
  shows "any_nstri_efficient_m f ss m = any_nstri_efficient_m g ss m"
using assms proof (induct ss arbitrary: m, simp)
  case (Cons s ss)
  show ?case 
  proof (cases "f s m")
    case (Pair res m')
    with Cons have rec: "any_nstri_efficient_m f ss m' = any_nstri_efficient_m g ss m'" by auto
    from Pair rec show ?thesis 
      by (simp add: Cons(2)[of s m, simplified], cases res, auto)    
  qed
qed

lemma all_stri_efficient_m [fundef_cong] : 
  assumes "\<And> t m. t \<in> set ts \<Longrightarrow> f t m = g t m"
  shows "all_stri_efficient_m f ts m = all_stri_efficient_m g ts m"
using assms proof (induct ts arbitrary: m, simp)
  case (Cons t ts)
  show ?case 
  proof (cases "f t m")
    case (Pair res m')
    with Cons have rec: "all_stri_efficient_m f ts m' = all_stri_efficient_m g ts m'" by auto
    from Pair rec show ?thesis 
      by (simp add: Cons(2)[of t m, simplified], cases res, auto)    
  qed
qed

(** Efficient filter **)

fun filter_not_stri_efficient_m_aux :: "('a \<Rightarrow> ('m, bool \<times> bool) memo) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> ('m, 'a list) memo"
where
  "filter_not_stri_efficient_m_aux f []       bs m = (bs, m)" |
  "filter_not_stri_efficient_m_aux f (a # as) bs m = (
    let ((s, _), m') = f a m in
    if s
    then filter_not_stri_efficient_m_aux f as bs m'
    else filter_not_stri_efficient_m_aux f as (a # bs) m')"

definition filter_not_stri_efficient_m:: "('a \<Rightarrow> ('m, bool \<times> bool) memo) \<Rightarrow> 'a list \<Rightarrow> ('m, 'a list) memo"
where
  "filter_not_stri_efficient_m f as m \<equiv> let (ys, m') = filter_not_stri_efficient_m_aux f as [] m in (rev ys, m')"

(** congruence lemmas needed for rpo's termination proof **)

lemma filter_not_stri_efficient_m_aux[fundef_cong]: 
  assumes "\<And> a b m. a \<in> set as \<Longrightarrow> f a m = g a m"
  shows "filter_not_stri_efficient_m_aux f as bs m =
   filter_not_stri_efficient_m_aux g as bs m"
using assms proof (induct as arbitrary: bs m)
case Nil thus ?case by simp
next
case (Cons a as bs m)
  show ?case
  proof (cases "f a m")
  case (Pair sns m')
    obtain s ns where Psns: "sns = (s, ns)" by (cases sns) auto
    from Pair and Psns and Cons(2) have "g a m = ((s, ns), m')" by simp
    with Psns and Pair and Cons show ?thesis by simp
  qed
qed
    
lemma filter_not_stri_efficient_m[fundef_cong]: 
  assumes "\<And> a b m. a \<in> set as \<Longrightarrow> f a m = g a m"
  shows "filter_not_stri_efficient_m f as m =
   filter_not_stri_efficient_m g as m"
unfolding filter_not_stri_efficient_m_def
using assms and filter_not_stri_efficient_m_aux
by force

(** These lemmas are needed for the termination prover to be able
    to find a decreasing lexographic order for mul_ext_efficient_m. **)

lemma filter_not_stri_efficient_m_aux_sublist:
  "sublisteq (rev (fst (filter_not_stri_efficient_m_aux f as bs m))) ((rev bs) @ as)"
proof (induct as arbitrary: bs m)
  case Nil thus ?case by simp
next
  case (Cons a as)
  obtain sns m' where P1: "f a m = (sns, m')" by (cases "f a m") auto
  obtain s ns where P2: "sns = (s, ns)" by (cases sns) auto
  show ?case
  proof (cases s)
    case True
    with Cons [of "bs" m'] P1 P2
    have e1: "sublisteq (rev (fst (filter_not_stri_efficient_m_aux f (a # as) bs m))) ((rev bs) @ as)"
     by simp
    with list_emb_append_mono [OF sublisteq_refl [of "rev bs"] list_emb_Cons [OF sublisteq_refl [of as], of a]]
      and sublisteq_trans show ?thesis by simp
  next
    case False with Cons [of "a # bs" m'] P1 P2 show ?thesis by simp
  qed
qed

lemma filter_not_stri_efficient_m_sublist:
  "sublisteq (fst (filter_not_stri_efficient_m f as m)) as"
proof -
  obtain ys m' where Hrew: "filter_not_stri_efficient_m_aux f as [] m = (ys, m')"
    by (cases "filter_not_stri_efficient_m_aux f as [] m") auto
  with filter_not_stri_efficient_m_aux_sublist[of f as "[]" m]
  have "sublisteq (rev ys) as" by simp
  thus ?thesis using Hrew unfolding filter_not_stri_efficient_m_def by simp
qed

lemma filter_not_stri_efficient_m_submultiset:
  "multiset_of (fst (filter_not_stri_efficient_m f as m)) \<le> multiset_of as"
  using sublisteq_submultiset filter_not_stri_efficient_m_sublist by force

lemma filter_not_stri_efficient_m_subset:
  "set (fst (filter_not_stri_efficient_m f as m)) \<le> set as"
  using sublisteq_subset filter_not_stri_efficient_m_sublist by force

lemma length_filter_not_stri_efficient_m[simp]:
 "length (fst (filter_not_stri_efficient_m f as m)) \<le> length as"
  using list_emb_length filter_not_stri_efficient_m_sublist by force

fun diag_l :: "('a \<times> 'b) \<Rightarrow> (('a \<times> 'a) \<times> 'b)" where
  "diag_l (a, b) = ((a, a), b)"

(** Efficient version of lex_ext **)

fun lex_ext_efficient_m :: "('a \<Rightarrow> 'a \<Rightarrow> ('m, bool \<times> bool) memo) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> ('m, bool \<times> bool) memo"
where
  "lex_ext_efficient_m f []     []     m = ((False, True), m)" |
  "lex_ext_efficient_m f (_#_)  []     m = ((True, True), m)" |
  "lex_ext_efficient_m f []     (_#_)  m = ((False, False), m)" |
  "lex_ext_efficient_m f (a#as) (b#bs) m = (let ((s,ns),m') = f a b m in
     if s then ((True, True), m')
          else if ns then (lex_ext_efficient_m f as bs m')
                     else ((False, False), m'))"

(** congruence lemmas needed for rpo's termination proof **)

lemma lex_ext_efficient_m_cong[fundef_cong] : fixes f g ss ts m
  assumes "\<And> s t m. \<lbrakk>s \<in> set ss; t \<in> set ts\<rbrakk> \<Longrightarrow> f s t m = g s t m"
  shows "lex_ext_efficient_m f ss ts m = lex_ext_efficient_m g ss ts m"
using assms 
proof (induct ss arbitrary: ts m)
  case Nil
  thus ?case by (cases ts, auto)
next
  case (Cons s ss) note oCons = this
  show ?case 
  proof (cases ts)
    case Nil
    with Cons show ?thesis by auto
  next
    case (Cons t ts')
    with oCons have id: "g s t m = f s t m" by simp
    show ?thesis
    proof (cases "f s t m")
      case (Pair res m')
      from oCons Cons have id2: "lex_ext_efficient_m f ss ts' m' = lex_ext_efficient_m g ss ts' m'" by auto
      show ?thesis
        by (simp add: Pair Cons id, cases res, auto simp: id2)
    qed
  qed
qed

(** Efficient version of mul_ext **)

fun mul_ext_efficient_m :: "('a \<Rightarrow> 'a \<Rightarrow> ('m, bool \<times> bool) memo) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> ('m, bool \<times> bool) memo"
and ex_dom_efficient_m :: "('a \<Rightarrow> 'a \<Rightarrow> ('m, bool \<times> bool) memo) \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> ('m, bool \<times> bool) memo"
where
  "mul_ext_efficient_m f [] []       m = ((False, True), m)" |
  "mul_ext_efficient_m f [] _        m = ((False, False),m)" |
  "mul_ext_efficient_m f _  []       m = ((True, True), m)" |
  "mul_ext_efficient_m f xs (y # ys) m = ex_dom_efficient_m f y xs xs ys m" |
  "ex_dom_efficient_m f y []       _  _  m = ((False, False), m)" |
  "ex_dom_efficient_m f y (b # bs) xs ys m = (
    let ((s, ns), m') = f b y m in
    if s
    then let (ys', m2) = filter_not_stri_efficient_m (f b) ys m' in
         let ((_, ns1), m3) = mul_ext_efficient_m f (remove1 b xs) ys' m2 in
         if ns1 then ((True, True), m3)
                else ex_dom_efficient_m f y bs xs ys m3
    else if ns
         then let ((s1, ns1), m'') = mul_ext_efficient_m f (remove1 b xs) ys m' in
              if s1 then ((True, True), m'')
                    else let ((s2, ns2), m3) = ex_dom_efficient_m f y bs xs ys m'' in
                         if s2 then ((True, True), m3)
                               else ((False, ns1 \<or> ns2), m3)
         else ex_dom_efficient_m f y bs xs ys m')"


(** congruence lemmas needed for rpo's termination proof **)
lemma ex_dom_efficient_m_cong:
  assumes "\<forall>x. x \<in> set bs \<longrightarrow> x \<in> set xs"
  and "\<And> x x' m. x \<in> set xs \<Longrightarrow> x' \<in> set (y # ys) \<Longrightarrow> f x x' m = g x x' m"
  and "\<forall>as bs m. length bs \<le> length ys \<longrightarrow>
   (\<forall>x x' m. x \<in> set as \<longrightarrow> x' \<in> set bs \<longrightarrow> f x x' m = g x x' m) \<longrightarrow>
   mul_ext_efficient_m f as bs m = mul_ext_efficient_m g as bs m"
  shows "ex_dom_efficient_m f y bs xs ys m = ex_dom_efficient_m g y bs xs ys m"
using assms(1) proof (induct bs arbitrary: m)
case Nil show ?case by simp
next
case (Cons b bs)
  from Cons(2) have "\<forall>x. x \<in> set bs \<longrightarrow> x \<in> set xs" by simp
  note IH = Cons(1)[OF this]
  show ?case
  proof (cases "f b y m")
  case (Pair sns m') note Psm' = this
    show ?thesis
    proof (cases sns)
    case (Pair s ns) note Psns = this
    from Cons(2) have b_in_xs: "b \<in> set xs" by simp
    from assms(2)[OF this] and Psm' and Psns have Psns': "g b y m = ((s, ns), m')" by simp
    from assms(2)[OF b_in_xs] and filter_not_stri_efficient_m[of ys "f b"]
     have T: "\<forall>m. filter_not_stri_efficient_m (f b) ys m = filter_not_stri_efficient_m (g b) ys m"
     by simp
    {
      fix x x' m m'
      assume x_in_bxs:"x \<in> set (remove1 b xs)"
        and x'_in_bys: "x' \<in> set (fst (filter_not_stri_efficient_m (f b) ys m))"
      from x_in_bxs and set_mp[OF set_remove1_subset] have x_in_xs: "x \<in> set xs" by fast
      from x'_in_bys and filter_not_stri_efficient_m_subset
       have "x' \<in> set (y # ys)" by force
      from assms(2)[OF x_in_xs this] have "f x x' m' = g x x' m'" by simp
    }
    hence "\<forall>x x' m m'. x \<in> set (remove1 b xs) \<longrightarrow>
     x' \<in> set (fst (filter_not_stri_efficient_m (f b) ys m)) \<longrightarrow> f x x' m' = g x x' m'" by fast
    from this assms(3)[rule_format, OF length_filter_not_stri_efficient_m, of "remove1 b xs" "f b"]
     have "\<forall>m m'. mul_ext_efficient_m f (remove1 b xs) (fst (filter_not_stri_efficient_m (f b) ys m)) m'
     = mul_ext_efficient_m g (remove1 b xs) (fst (filter_not_stri_efficient_m (f b) ys m)) m'" by blast
    with T have Heq: "\<forall>m m'. mul_ext_efficient_m f (remove1 b xs)
     (fst (filter_not_stri_efficient_m (f b) ys m)) m' = mul_ext_efficient_m g (remove1 b xs)
     (fst (filter_not_stri_efficient_m (g b) ys m)) m'" by simp
    {
      fix x x' m
      assume x_in_bxs:"x \<in> set (remove1 b xs)"
        and x'_in_ys: "x' \<in> set ys"
      from x_in_bxs and set_mp[OF set_remove1_subset] have x_in_xs: "x \<in> set xs" by fast
      from x'_in_ys have "x' \<in> set (y # ys)" by simp
      from assms(2)[OF x_in_xs this] have "f x x' m = g x x' m" by assumption
    }
    hence "\<forall>x x' m. x \<in> set (remove1 b xs) \<longrightarrow> x' \<in> set ys \<longrightarrow> f x x' m = g x x' m" by fast
    with assms(3) have Heq2: "\<forall>m. mul_ext_efficient_m f (remove1 b xs) ys m =
     mul_ext_efficient_m g (remove1 b xs) ys m" by simp
    obtain ed1 ed2 where Hed: "filter_not_stri_efficient_m (f b) ys m' = (ed1, ed2)"
     by (cases "filter_not_stri_efficient_m (f b) ys m'") auto
    with T have Hed': "filter_not_stri_efficient_m (g b) ys m' = (ed1, ed2)" by simp
    obtain me12 m3 where Hme: "mul_ext_efficient_m f (remove1 b xs) ed1 ed2 = (me12, m3)"
     by (cases "mul_ext_efficient_m f (remove1 b xs) ed1 ed2") auto
    with Heq Hed Hed' have Hme': "mul_ext_efficient_m g (remove1 b xs) ed1 ed2 = (me12, m3)"
     by (metis fst_conv)
    obtain me34 m4 where Hme2: "mul_ext_efficient_m g (remove1 b xs) ys m' = (me34, m4)"
     by (cases "mul_ext_efficient_m g (remove1 b xs) ys m'") auto 
    with Heq2 have Hme2': "mul_ext_efficient_m f (remove1 b xs) ys m' = (me34, m4)" by simp
    obtain me1 me2 where Hme12: "me12 = (me1, me2)" by (cases me12) auto
    obtain me3 me4 where Hme34: "me34 = (me3, me4)" by (cases me34) auto
    from Hed Hed' Hme Hme2 Hme' Hme2' Hme12 Hme34 T Heq Heq2 IH Psns Psns' Psm'
     show ?thesis by simp
    qed
  qed
qed

lemma mul_ext_efficient_m_cong_prelim:
  "mul_ext_efficient_m f xs [] m = mul_ext_efficient_m g xs [] m"
by (cases xs, simp_all)

lemma mul_ext_efficient_m_cong_ind:
  assumes "\<And> x x' m. x \<in> set xs \<Longrightarrow> x' \<in> set ys \<Longrightarrow> f x x' m = g x x' m"
  and "length ys \<le> n"
  shows "mul_ext_efficient_m f xs ys m = mul_ext_efficient_m g xs ys m"
using assms proof (induct n arbitrary: xs ys m)
case (0 xs ys)
  hence "ys = []" by simp
  with 0(1) and mul_ext_efficient_m_cong_prelim show ?case by simp
next
case (Suc n xs ys)
  show ?case
  proof (cases "ys")
  case Nil with Suc(2) and mul_ext_efficient_m_cong_prelim show ?thesis by simp
  next
  case (Cons t ts) note Htts = this
    show ?thesis
    proof (cases xs)
    case Nil with Htts show ?thesis by simp
    next
    case (Cons s ss)
      with Htts have Hrew: "\<And>f m. mul_ext_efficient_m f xs ys m = ex_dom_efficient_m f t xs xs ts m" by simp
      have xs_in_xs: "\<forall>x. x \<in> set xs \<longrightarrow> x \<in> set xs" by simp
      from Suc(2) and Htts
       have fg_ext: "\<And>x x' m. \<lbrakk>x \<in> set xs; x' \<in> set (t # ts)\<rbrakk> \<Longrightarrow> f x x' m = g x x' m" by simp
      {
        fix as::"'a list" and bs::"'a list"  and m
        assume lbs_bd: "length bs \<le> length ts"
         and local_fg_ext: "\<forall>x x' m. x \<in> set as \<longrightarrow> x' \<in> set bs \<longrightarrow> f x x' m = g x x' m"
        from Suc(3) and Htts and lbs_bd have "length bs \<le> n" by simp
        from Suc(1)[OF _ this] and local_fg_ext
        have "mul_ext_efficient_m f as bs m = mul_ext_efficient_m g as bs m" by simp
      }
      hence T: "\<forall>as bs m. length bs \<le> length ts \<longrightarrow> (\<forall>x x' m. x \<in> set as \<longrightarrow> x' \<in> set bs \<longrightarrow>
       f x x' m = g x x' m) \<longrightarrow> mul_ext_efficient_m f as bs m = mul_ext_efficient_m g as bs m" by simp
      from ex_dom_efficient_m_cong[OF xs_in_xs, of t ts f g m] fg_ext T
       have "ex_dom_efficient_m f t xs xs ts m = ex_dom_efficient_m g t xs xs ts m" by simp
      thus ?thesis using Htts and Cons by simp
    qed
  qed
qed

lemma mul_ext_efficient_m_cong[fundef_cong]:
  assumes "\<And> x x' m. x \<in> set xs \<Longrightarrow> x' \<in> set ys \<Longrightarrow> f x x' m = g x x' m"
  shows "mul_ext_efficient_m f xs ys m = mul_ext_efficient_m g xs ys m"
using mul_ext_efficient_m_cong_ind[OF assms(1)] by fast

(** Efficient version of rpo that uses a memory. *)

fun rpo_efficient_m ::
 "('m, ('f,'v) term \<times> ('f, 'v) term, bool \<times> bool) memory \<Rightarrow> 'f prec_fct \<Rightarrow> 'f status_fct \<Rightarrow>
  ('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> ('m, bool \<times> bool) memo" where
 "rpo_efficient_m model pr tag s t m = (case (memory.lookup model m (s,t)) of
      Some v \<Rightarrow> (v, m)
    | None \<Rightarrow> let (res, res_m) =
    (case (s,t) of
       (Var x, Var y) \<Rightarrow> ((False, x = y), m)
     | (Var x, Fun g ts) \<Rightarrow> ((False, ts = [] \<and> (snd pr) (g,0)), m)
     | (Fun f ss, Var y) \<Rightarrow> diag_l (any_nstri_efficient_m (\<lambda>s. rpo_efficient_m model pr tag s (Var y)) ss m)
     | (Fun f ss, Fun g ts) \<Rightarrow>
     (let (any, m') = any_nstri_efficient_m (\<lambda>s. rpo_efficient_m model pr tag s (Fun g ts)) ss m in
      if any then ((True, True), m')
             else let m = length ss in let n = length ts in
                  let (prs, prns) = fst pr (f,m) (g,n) in
                  if prns
                  then let (all, m'') = all_stri_efficient_m (rpo_efficient_m model pr tag (Fun f ss)) ts m' in
                       if all
                       then if prs
                            then ((True, True), m'')
                            else let cf = tag (f,m) in
                                 if cf = tag (g,n)
                                 then if cf = Mul
                                      then mul_ext_efficient_m (rpo_efficient_m model pr tag) ss ts m''
                                      else lex_ext_efficient_m (rpo_efficient_m model pr tag) ss ts m''
                                 else let b = (n = 0) in ((m \<noteq> 0 \<and> b, b), m'')
                       else ((False, False), m'')
                  else ((False, False), m')
    )) in (res, memory.store model res_m ((s, t), res)))"


(** Now that all the functions are defined and proved terminating,
    we prove their soundness. This soundness is obviously true iff
    the provided memory is satisfactory. *)

locale memoized_rpo =
  fixes model :: "('m, ('f,'v) term \<times> ('f, 'v) term, bool \<times> bool)memory"
  fixes model_P  :: "('m, ('f, 'v)term \<times> ('f, 'v)term, bool \<times> bool)memory \<Rightarrow> 'm
                    \<Rightarrow> ((('f, 'v)term \<times> ('f, 'v)term) \<Rightarrow> bool \<times> bool) \<Rightarrow> bool"
  assumes model_satisfactory: "satisfactory model_P model"
begin

abbreviation rpo_efficient where "rpo_efficient \<equiv> rpo_efficient_m model"
definition m_rpo_valid where
  "m_rpo_valid pr tag m = model_P model m (\<lambda>p. rpo_unbounded_impl pr tag (fst p) (snd p))"

lemma lookup_valid:
  assumes "model_P model m f"
  shows "valid model m f"
using model_satisfactory assms by simp

lemma store_valid:
  assumes "model_P model M f"
  shows "model_P model (memory.store model M (a, f a)) f"
using model_satisfactory assms by (cases a) auto

(** Our auxiliary functions are sound. *)

lemma any_nstri_efficient_m_sound:
  assumes "\<And>s m. s \<in> set ss \<and> m_valid f m \<Longrightarrow> fst (fE s m) = f s t \<and> m_valid f (snd (fE s m))"
  and "m_valid f m"
  shows "m_valid f (snd (any_nstri_efficient_m fE ss m))
         \<and> fst (any_nstri_efficient_m fE ss m) = (\<exists> s \<in> set ss. snd (f s t))"
using assms proof (induct ss arbitrary: m)
case Nil
  have "any_nstri_efficient_m fE [] m = (False,m)" by simp
  moreover have "(\<exists> s \<in> set []. snd (f s t)) = False" by simp
  ultimately show ?case using Nil(2) by simp
next
case (Cons u us)
  obtain v m' where fE_val: "fE u m = (v, m')" by (cases "fE u m") auto
  hence m':"snd (fE u m) = m'" by simp
  from Cons have u_vals_sound:"fst (fE u m) = f u t \<and> m_valid f (snd (fE u m))" by simp
  with m' have m'_valid:"m_valid f m'" by (simp add: assms)
  obtain s ns where v_def: "v = (s, ns)" by (cases v) auto
  with u_vals_sound fE_val have f_val:"f u t = (s, ns)" by simp
  show ?case
  proof (cases ns)
  case True
    with v_def fE_val have "snd (any_nstri_efficient_m fE (u # us) m) = m'" by simp
    with m'_valid have valid_memory: "m_valid f (snd (any_nstri_efficient_m fE (u # us) m))" by simp
    from True v_def fE_val have any_nstri_efficient_m_val:
     "fst (any_nstri_efficient_m fE (u # us) m) = True" by simp
    from True f_val have ex_val:"(\<exists> s \<in> set (u # us). snd (f s t)) = True" by simp
    with any_nstri_efficient_m_val valid_memory show ?thesis by simp
  next
  case False
    with v_def fE_val have any_nstri_efficient_m_val:
     "any_nstri_efficient_m fE (u # us) m = any_nstri_efficient_m fE us m'" by simp
    from Cons(2) have "\<And>s m. s \<in> set us \<and> m_valid f m \<Longrightarrow> fst (fE s m) = f s t
     \<and> m_valid f (snd (fE s m))" by simp
    with Cons(1)[of m'] m'_valid have rec:"m_valid f (snd (any_nstri_efficient_m fE us m'))
     \<and> fst (any_nstri_efficient_m fE us m') = (\<exists> s \<in> set us. snd (f s t))" by simp
    from f_val False have ex_val:"(\<exists> s \<in> set (u # us). snd (f s t)) = (\<exists>s \<in> set us. snd (f s t))"
     by simp
    with rec any_nstri_efficient_m_val show ?thesis by simp
  qed
qed

lemma all_stri_efficient_m_sound:
  assumes "\<And>t m. t \<in> set ts \<and> m_valid f m \<Longrightarrow> fst (fE t m) = f s t \<and> m_valid f (snd (fE t m))"
  and "m_valid f m"
  shows "m_valid f (snd (all_stri_efficient_m fE ts m))
   \<and> fst (all_stri_efficient_m fE ts m) = (\<forall> t \<in> set ts. fst (f s t))"
using assms proof (induct ts arbitrary: m)
case Nil
  have "all_stri_efficient_m fE [] m = (True, m)" by simp
  moreover have "(\<forall> t \<in> set []. fst (f s t)) = True" by simp
  ultimately show ?case using Nil(2) by simp
next
case (Cons u us)
  obtain v m' where fE_val: "fE u m = (v, m')" by (cases "fE u m") auto
  hence m':"snd (fE u m) = m'" by simp
  from Cons have u_vals_sound:"fst (fE u m) = f s u \<and> m_valid f (snd (fE u m))" by simp
  with m' have m'_valid: "m_valid f m'" by (simp add: assms)
  obtain stri nstri where v_def: "v = (stri, nstri)" by (cases v) auto
  with u_vals_sound fE_val have f_val:"f s u = (stri, nstri)" by simp
  show ?case
  proof (cases stri)
  case False
    with v_def fE_val have "snd (all_stri_efficient_m fE (u # us) m) = m'" by simp
    with m'_valid have valid_memory:"m_valid f (snd (all_stri_efficient_m fE (u # us) m))" by simp
    from False v_def fE_val have all_stri_efficient_m_val:
     "fst (all_stri_efficient_m fE (u # us) m) = False" by simp
    with False f_val valid_memory show ?thesis by simp
  next
  case True
    with v_def fE_val have all_stri_efficient_m_val:
     "all_stri_efficient_m fE (u # us) m = all_stri_efficient_m fE us m'" by simp
    from Cons(2) have "\<And>t m. t \<in> set us \<and> m_valid f m \<Longrightarrow> fst (fE t m) = f s t
     \<and> m_valid f (snd (fE t m))" by simp
    with Cons(1)[of m'] m'_valid have rec:"m_valid f (snd (all_stri_efficient_m fE us m'))
     \<and> fst (all_stri_efficient_m fE us m') = (\<forall> t \<in> set us. fst (f s t))" by simp
    from f_val True have all_val:"(\<forall> t \<in> set (u # us). fst (f s t)) = (\<forall> t \<in> set us. fst (f s t))"
     by simp
    with rec all_stri_efficient_m_val show ?thesis by simp
  qed
qed

lemma filter_not_stri_efficient_m_aux_sound:
  assumes "\<And>t m. t \<in> set ts \<and> model_P model m f \<Longrightarrow> fst (fE t m) = f (s, t) \<and> model_P model (snd (fE t m)) f"
  and "model_P model m f"
  and "\<forall> t \<in> set res. \<not> fst (f (s, t))"
  shows "model_P model (snd (filter_not_stri_efficient_m_aux fE ts res m)) f
   \<and> rev (fst (filter_not_stri_efficient_m_aux fE ts res m)) = [t\<leftarrow> ((rev res) @ ts) . \<not> fst (f (s, t))]"
using assms proof (induct ts arbitrary: m res)
case Nil show ?case
  proof (rule conjI)
    from Nil(2) show "model_P model (snd (filter_not_stri_efficient_m_aux fE [] res m)) f" by simp
    from Nil(3) show "rev (fst (filter_not_stri_efficient_m_aux fE [] res m))
     = [t\<leftarrow> (rev res @ []) . \<not> fst (f (s, t))]" by (induct res, simp_all)
  qed
next
case (Cons u us)
  obtain v m' where fE_val: "fE u m = (v, m')" by (cases "fE u m") auto
  hence m':"snd (fE u m) = m'" by simp
  from Cons have u_vals_sound:"fst (fE u m) = f (s, u) \<and> model_P model (snd (fE u m)) f" by simp
  with m' have m'_valid: "model_P model m' f" by (simp add: assms)
  obtain stri nstri where v_def: "v = (stri, nstri)" by (cases v) auto
  with u_vals_sound fE_val have f_val:"f (s, u) = (stri, nstri)" by simp
  show ?case
  proof (cases stri)
  case True
    with v_def fE_val
     have all_stri_efficient_m_val: "filter_not_stri_efficient_m_aux fE (u # us) res m
     = filter_not_stri_efficient_m_aux fE us res m'" by simp
    from f_val True have nfsu: "(\<not> fst (f (s, u))) = False" by simp
    from Cons(2) have "\<And>t m. t \<in> set us \<and> model_P model m f \<Longrightarrow> fst (fE t m) = f (s, t)
     \<and> model_P model (snd (fE t m)) f" by simp
    with Cons(1)[of m' res] Cons(4) m'_valid all_stri_efficient_m_val nfsu
     show ?thesis by simp
  next
  case False
    with v_def fE_val
     have all_stri_efficient_m_val: "filter_not_stri_efficient_m_aux fE (u # us) res m
      = filter_not_stri_efficient_m_aux fE us (u # res) m'" by simp
    from Cons(4) False f_val have acc_sound: "\<forall>t \<in> set (u # res). \<not> fst (f (s, t))" by simp
    from Cons(2) have "\<And>t m. t \<in> set us \<and> model_P model m f \<Longrightarrow> fst (fE t m) = f (s, t)
     \<and> model_P model (snd (fE t m)) f" by simp
    with Cons(1)[of m' "u # res"] acc_sound m'_valid all_stri_efficient_m_val
     show ?thesis by simp
  qed
qed

lemma filter_not_stri_efficient_m_sound:
  assumes "\<And>t m. t \<in> set ts \<and> model_P model m f \<Longrightarrow> fst (fE t m) = f (s, t) \<and> model_P model (snd (fE t m)) f"
  and "model_P model m f"
  shows "model_P model (snd (filter_not_stri_efficient_m fE ts m)) f
   \<and> fst (filter_not_stri_efficient_m fE ts m) = [y\<leftarrow>ts . \<not> fst (f (s, y))]"
proof -
obtain ys m' where Hrew: "filter_not_stri_efficient_m_aux fE ts [] m = (ys, m')"
 by (cases "filter_not_stri_efficient_m_aux fE ts [] m") auto
from filter_not_stri_efficient_m_aux_sound[OF assms(1) assms(2), of ts "\<lambda>_ m. m" "[]"] assms Hrew
show ?thesis unfolding filter_not_stri_efficient_m_def by simp
qed

(** lex_ext_efficient_m is sound **)

declare lex_ext_unbounded.simps[simp]

lemma lex_ext_efficient_m_sound:
  assumes "\<And>t s m.  s \<in> set ss \<and> t \<in> set ts \<and> model_P model m f \<Longrightarrow> fst (fE s t m) = f (s, t)
   \<and> model_P model (snd (fE s t m)) f"
  and "model_P model m f"
  shows "model_P model (snd (lex_ext_efficient_m fE ss ts m)) f
   \<and> fst (lex_ext_efficient_m fE ss ts m) = lex_ext_unbounded (\<lambda>s t. f (s, t)) ss ts"
using assms proof (induct ss ts arbitrary: m rule: List.list_induct2')
case (1 m)
  have "lex_ext_efficient_m fE [] [] m = ((False,True),m)" by simp
  moreover have "lex_ext_unbounded (\<lambda> s t. f (s, t)) [] [] = (False, True)" by simp
  ultimately show ?case using 1(2) by simp
next
case (2 a as m)
  have "lex_ext_efficient_m fE (a # as) [] m = ((True,True),m)" by simp
  moreover have "lex_ext_unbounded (\<lambda>s t. f (s, t)) (a # as) [] = (True, True)" by simp
  ultimately show ?case using 2(2) by simp
next
case (3 b bs m)
  have "lex_ext_efficient_m fE [] (b # bs) m = ((False,False),m)" by simp
  moreover have "lex_ext_unbounded (\<lambda>s t. f (s, t)) [] (b # bs)  = (False, False)" by simp
  ultimately show ?case using 3(2) by simp
next
case (4 a as b bs m)
  obtain v m' where fE_val: "fE a b m = (v, m')" by (cases "fE a b m") auto
  from 4(2)[of a b m] 4(3) have fE_sound: "fst (fE a b m) = f (a, b) \<and> model_P model (snd (fE a b m)) f"
   by simp
  with fE_val have m'_valid: "model_P model m' f" by simp
  obtain s ns where v_def: "v = (s, ns)" by (cases v) auto
  with fE_sound fE_val have f_val:"f (a, b) = (s,ns)" by simp
  show ?case
  proof (cases s)
  case True
    with fE_val v_def have lex_ext_efficient_m_val:
     "lex_ext_efficient_m fE (a # as) (b # bs) m = ((True,True),m')" by simp
    from f_val True
     have lex_ext_unbounded_val:"lex_ext_unbounded (\<lambda>s t. f (s, t)) (a # as) (b # bs) = (True,True)"
     by simp
    with m'_valid lex_ext_efficient_m_val show ?thesis by simp
  next
  case False note sFalse = this
    show ?thesis
    proof (cases ns)
    case True
      from 4(2) have "\<And>s t m. s \<in> set as \<and> t \<in> set bs \<and> model_P model m f \<Longrightarrow> fst (fE s t m) = f (s, t)
       \<and> model_P model (snd (fE s t m)) f" by simp
      from this 4(1)[of m'] m'_valid have rec: "model_P model (snd (lex_ext_efficient_m fE as bs m')) f
       \<and> fst (lex_ext_efficient_m fE as bs m') = lex_ext_unbounded (\<lambda>s t. f (s, t)) as bs"
       by simp
      from sFalse True v_def fE_val
       have lex_ext_efficient_m_val: "lex_ext_efficient_m fE (a # as) (b # bs) m
       = lex_ext_efficient_m fE  as bs m'" by simp
      from sFalse True f_val
       have lex_ext_unbounded_val:"lex_ext_unbounded (\<lambda>s t. f (s, t)) (a # as) (b # bs)
       = lex_ext_unbounded (\<lambda>s t. f (s, t)) as bs" by simp
      with lex_ext_efficient_m_val rec show ?thesis by simp
    next
    case False
      with sFalse v_def fE_val have lex_ext_efficient_m_val:
       "lex_ext_efficient_m fE (a # as) (b # bs) m = ((False,False),m')" by simp
      from sFalse False f_val 
       have lex_ext_unbounded_val: "lex_ext_unbounded (\<lambda>s t. f (s, t)) (a # as) (b # bs) = (False,False)"
       by simp
      with lex_ext_efficient_m_val m'_valid show ?thesis by simp
    qed
  qed
qed

declare lex_ext_unbounded.simps[simp del]

(** ex_dom_efficient_m is sound **)

lemma ex_dom_efficient_m_sound:
  assumes "\<And>t s m.  s \<in> set ss \<and> t \<in> set (y # ts) \<and> model_P model m f \<Longrightarrow> fst (fE s t m) = f (s, t)
   \<and> model_P model (snd (fE s t m)) f"
  and "\<forall>x. x \<in> set bs \<longrightarrow> x \<in> set ss"
  and "\<forall>as bs m. sublisteq bs ts \<longrightarrow> sublisteq as ss \<longrightarrow> model_P model m f \<longrightarrow>
   model_P model (snd (mul_ext_efficient_m fE as bs m)) f
   \<and> fst (mul_ext_efficient_m fE as bs m) = mul_ext_impl (\<lambda>s t. f (s, t)) as bs"
  and "model_P model m f"
  shows "model_P model (snd (ex_dom_efficient_m fE y bs ss ts m)) f
   \<and> fst (ex_dom_efficient_m fE y bs ss ts m) = ex_dominating (\<lambda>s t. f (s, t)) y bs ss ts"
using assms(2) assms(4) proof (induct bs arbitrary: m)
case Nil thus ?case by simp
next
case (Cons b bs)
  from Cons(2) have inc_rec: "\<forall>x. x \<in> set bs \<longrightarrow> x \<in> set ss" by simp
  from Cons(2) have b_in_ss: "b \<in> set ss" by simp
  obtain v m' where fE_val: "fE b y m = (v, m')" by (cases "fE b y m") auto
  from assms(1) Cons(2)[rule_format, of b] Cons(3)
   have fE_sound: "fst (fE b y m) = f (b, y) \<and> model_P model (snd (fE b y m)) f" by simp
  with fE_val have m'_valid: "model_P model m' f" by simp
  obtain s ns where v_def: "v = (s, ns)" by (cases v) auto
  with fE_sound fE_val have f_val:"f (b, y) = (s,ns)" by simp
  thus ?case 
  proof (cases s)
  case True note Hs = this
    obtain ys' m2 where filt: "filter_not_stri_efficient_m (fE b) ts m' = (ys', m2)"
     by (cases "filter_not_stri_efficient_m (fE b) ts m'") auto
    from assms(1)[of b] b_in_ss have T: "\<And>t m. t \<in> set ts \<and> model_P model m f \<Longrightarrow> fst (fE b t m) = f (b, t)
     \<and> model_P model (snd (fE b t m)) f" by simp
    from filter_not_stri_efficient_m_sound[OF T m'_valid] filt
     have filt': "ys' = [y\<leftarrow>ts. \<not> fst (f (b, y))]" by force
    hence filt'': "multiset_of ys' = multiset_of [y\<leftarrow>ts. \<not> fst (f (b, y))]" by simp
    from filter_not_stri_efficient_m_sound[OF T m'_valid] filt
     have m2_valid: "model_P model m2 f" by force
    obtain v' m3 where mext: "mul_ext_efficient_m fE (remove1 b ss) ys' m2 = (v', m3)"
      by (cases "mul_ext_efficient_m fE (remove1 b ss) ys' m2") auto
    obtain s1 ns1 where v'_def: "v' = (s1, ns1)" by (cases v') auto
    from assms(3)[rule_format, of ys' "remove1 b ss" m2, OF _ _ m2_valid] filt
     filter_not_stri_efficient_m_sublist[of "fE b" ts m']
     remove1_sublisteq[of b ss]
     have IH: "model_P model (snd (mul_ext_efficient_m fE (remove1 b ss) ys' m2)) f \<and>
     fst (mul_ext_efficient_m fE (remove1 b ss) ys' m2) = mul_ext_impl (\<lambda>s t. f (s, t)) (remove1 b ss) ys'"
     by simp
    from mul_ext_cong[OF _ filt'']
     have "mul_ext (\<lambda>s t. f (s, t)) (remove1 b ss) ys'
     = mul_ext (\<lambda>s t. f (s, t)) (remove1 b ss) [y\<leftarrow>ts . \<not> fst (f (b, y))]" by simp
    with mul_ext_impl_sound[of _ "remove1 b ss"] IH mext v'_def
     have mext': "mul_ext_impl (\<lambda>s t. f (s, t)) (remove1 b ss) [y\<leftarrow>ts . \<not> fst (f (b, y))] = (s1, ns1)" by simp
    from IH mext have m3_valid: "model_P model m3 f" by simp
    show ?thesis
    proof (cases ns1)
    case True note Hns1 = this
      thus ?thesis using f_val fE_val v_def v'_def Hns1 Hs filt filt' mext mext' m3_valid by simp
    next
    case False note Hns1 = this
      from Cons(1)[OF inc_rec m3_valid] f_val fE_val v_def v'_def Hns1 Hs filt filt' mext mext'
       show ?thesis by simp
    qed
  next
  case False note Hs = this
    thus ?thesis
    proof (cases ns)
    case True note Hns = this
      obtain v' m2 where mext: "mul_ext_efficient_m fE (remove1 b ss) ts m' = (v', m2)"
       by (cases "mul_ext_efficient_m fE (remove1 b ss) ts m'") auto
      with assms(3)[rule_format, of ts "remove1 b ss" m', OF _ _ m'_valid]
       remove1_sublisteq[of b ss]
       have m2_valid: "model_P model m2 f" by simp
      obtain s1 ns1 where v'_def: "v' = (s1, ns1)" by (cases v') auto
      from mext v'_def assms(3)[rule_format, of ts "remove1 b ss" m', OF _ _ m'_valid]
       remove1_sublisteq[of b ss]
       have mext': "mul_ext_impl (\<lambda>s t. f (s, t)) (remove1 b ss) ts = v'" by simp
      show ?thesis
      proof (cases s1)
      case True note Hs1 = this
        from mul_ext_impl_sound[of _ "remove1 b ss"] mul_ext_stri_imp_nstri
         have s_ns: "fst (mul_ext_impl (\<lambda>s t. f (s, t)) (remove1 b ss) ts)
         \<longrightarrow> snd (mul_ext_impl (\<lambda>s t. f (s, t)) (remove1 b ss) ts)"
         by force
        from mext' v'_def Hs1 s_ns have mext': "mul_ext_impl (\<lambda>s t. f (s, t)) (remove1 b ss) ts = (True, True)"
         by simp
        thus ?thesis using f_val fE_val v_def v'_def Hs Hs1 Hns mext mext' m2_valid by simp
      next
      case False note Hs1 = this
        obtain v2 m3 where mexd: "ex_dom_efficient_m fE y bs ss ts m2 = (v2, m3)"
         by (cases "ex_dom_efficient_m fE y bs ss ts m2") auto
        with Cons(1)[OF inc_rec m2_valid]
         have m3_valid: "model_P model m3 f" and mexd': "ex_dominating (\<lambda>s t. f (s, t)) y bs ss ts = v2"
         by auto
        obtain s2 ns2 where v2_def: "v2 = (s2, ns2)" by (cases v2) auto
        with ex_dominating_sound[OF inc_rec] 
         cond1_cond2 mexd' have "s2 \<longrightarrow> ns2" unfolding mul_ext_impl_sound by force
        thus ?thesis using f_val fE_val v_def v'_def v2_def Hs Hs1 Hns mext mext' mexd mexd' m3_valid
         by simp
      qed
    next
    case False note Hns = this
      show ?thesis using f_val fE_val v_def Hs Hns Cons(1)[OF inc_rec m'_valid] by simp
    qed
  qed
qed

(** mul_ext_efficient_m is sound *)

lemma mul_ext_efficient_m_sound_ind:
  assumes "\<And>t s m.  s \<in> set ss \<and> t \<in> set ts \<and> model_P model m f \<Longrightarrow> fst (fE s t m) = f (s, t)
   \<and> model_P model (snd (fE s t m)) f"
  and "model_P model m f"
  and "sublisteq ps ss"
  and "sublisteq qs ts"
  shows "model_P model (snd (mul_ext_efficient_m fE ps qs m)) f
   \<and> fst (mul_ext_efficient_m fE ps qs m) = mul_ext_impl (\<lambda>s t. f (s, t)) ps qs"
using assms(1) assms(2) assms(3) assms(4)
proof (induct ts arbitrary: ps qs m)
  case Nil
  from Nil(4) list_emb_Nil2 have "qs = []" by fast
  with Nil(2) show ?case by (cases ps, simp_all)
next
  case (Cons t ts)
  note IH = this
  note m_valid = Cons(3)
  from sublisteq_subset[OF IH(4)] have in_ps_ss: "\<And>s. s \<in> set ps \<Longrightarrow> s \<in> set ss" by fast
  from sublisteq_subset[OF IH(5)] have in_qs_ts: "\<And>s. s \<in> set qs \<Longrightarrow> s \<in> set (t # ts)" by fast
  show ?case using IH(5)
  proof (cases rule: list_emb.cases [of "op ="])
    case list_emb_Nil
    with IH show ?thesis by simp
  next
    case list_emb_Cons
    from IH(1) [OF _ m_valid IH(4) list_emb_Cons] IH(2) show ?thesis by simp
  next
    case (list_emb_Cons2 _ qs)
    note Htqs = list_emb_Cons2(1,2)
    show ?thesis
    proof (cases ps)
      case Nil thus ?thesis using Htqs m_valid by simp
    next
      case (Cons p ps) note Hpps = this
      obtain v m' where fE_val: "fE p t m = (v, m')" by (cases "fE p t m") auto
      from fE_val IH(2)[of p t m] in_ps_ss in_qs_ts m_valid Hpps Htqs have f_val: "f (p, t) = v" by simp
      from fE_val IH(2)[of p t m] in_ps_ss in_qs_ts m_valid Hpps Htqs
       have m'_valid: "model_P model m' f" by simp
      obtain s ns where v_def: "v = (s, ns)" by (cases v) auto
      {
        fix as bs m
        assume sub_bs_qs: "sublisteq bs qs" and sub_as_pps: "sublisteq as (p # ps)"
         and m_valid: "model_P model m f"
        from sublisteq_trans[OF sub_as_pps] IH(4) Hpps have sub_as_ss: "sublisteq as ss" by simp
        note sub_bs_ts = sublisteq_trans[OF sub_bs_qs list_emb_Cons2(3)]
        from IH(1)[OF _ m_valid sub_as_ss sub_bs_ts] IH(2)
         have " model_P model (snd (mul_ext_efficient_m fE as bs m)) f \<and>
         fst (mul_ext_efficient_m fE as bs m) = mul_ext_impl (\<lambda>s t. f (s, t)) as bs" by simp
      }
      hence rec: "\<forall>as bs m. sublisteq bs qs \<longrightarrow> sublisteq as (p # ps) \<longrightarrow> model_P model m f \<longrightarrow>
       model_P model (snd (mul_ext_efficient_m fE as bs m)) f \<and>
       fst (mul_ext_efficient_m fE as bs m) = mul_ext_impl (\<lambda>s t. f (s, t)) as bs" by simp
      from sublisteq_subset[OF IH(4)] Hpps have in_pps_ss: "\<forall>x. x \<in> set (p # ps) \<longrightarrow> x \<in> set ss" by fast
      from sublisteq_subset[OF IH(5)] Htqs have in_tqs_tts: "\<forall>x. x \<in> set (t # qs) \<longrightarrow> x \<in> set (t # ts)"
       by fast
      from IH(2) in_pps_ss in_tqs_tts 
       have IH2': "\<And>a b m. a \<in> set (p # ps) \<and> b \<in> set (t # qs) \<and> model_P model m f \<Longrightarrow>
       fst (fE a b m) = f (a, b) \<and> model_P model (snd (fE a b m)) f" by fast
      have ps_in_pps: "\<forall>x. x \<in> set ps \<longrightarrow> x \<in> set (p # ps)" by simp
      show ?thesis
      proof (cases ns)
      case True note Hns = this
        show ?thesis
        proof (cases s)
        case True note Hs = this
          obtain ys m2 where filt: "filter_not_stri_efficient_m (fE p) qs m' = (ys, m2)"
           by (cases "filter_not_stri_efficient_m (fE p) qs m'") auto
          from filt filter_not_stri_efficient_m_sublist[of "fE p" qs m']
           have sub_ys_qs: "sublisteq ys qs" by simp
          from filt filter_not_stri_efficient_m_sound[OF _ m'_valid, of qs "fE p"]
           IH(2)[of p] in_ps_ss in_qs_ts Htqs Hpps
           have m2_valid: "model_P model m2 f" and filt': "ys = [y\<leftarrow> qs. \<not> fst (f (p, y))]" by auto
          obtain v' m3 where mext: "mul_ext_efficient_m fE ps ys m2 = (v', m3)"
           by (cases "mul_ext_efficient_m fE ps ys m2") auto
          from IH(4) Hpps sublisteq_trans list_emb.list_emb_Cons [of "op =" ps ps p] have sub_ps_ss: "sublisteq ps ss" by simp
          from IH(1)[OF _ m2_valid sub_ps_ss sublisteq_trans[OF sub_ys_qs list_emb_Cons2(3)]] mext filt'
           IH(2) in_ps_ss in_qs_ts Htqs Hpps
           have mext': "mul_ext_impl (\<lambda>s t. f (s, t)) ps [y\<leftarrow> qs. \<not> fst (f (p, y))] = v'"
           and m3_valid: "model_P model m3 f"
           by auto
          obtain s' ns' where v'_def: "v' = (s', ns')" by (cases v') auto
          show ?thesis
          proof (cases ns')
          case True note Hns' = this
            from Htqs Hpps fE_val f_val v_def v'_def Hns Hns' Hs filt mext mext' m3_valid
             show ?thesis by simp
          next
          case False note Hns' = this
            from IH2' ex_dom_efficient_m_sound[OF _ ps_in_pps rec m3_valid, of t]
             Htqs Hpps fE_val f_val v_def v'_def Hns Hns' Hs filt mext mext'
             show ?thesis by simp
          qed
        next
        case False note Hs = this
          obtain v' m2 where mext: "mul_ext_efficient_m fE ps qs m' = (v', m2)"
           by (cases "mul_ext_efficient_m fE ps qs m'") auto
          from IH(4) Hpps sublisteq_trans list_emb.list_emb_Cons [of "op =" ps ps p] have sub_ps_ss: "sublisteq ps ss" by simp
          from IH(1)[OF _ m'_valid sub_ps_ss list_emb_Cons2(3)] mext IH(2) in_ps_ss in_qs_ts Htqs Hpps
           have mext': "mul_ext_impl (\<lambda>s t. f (s, t)) ps qs = v'" and m2_valid: "model_P model m2 f" by auto
          obtain s' ns' where v'_def: "v' = (s', ns')" by (cases v') auto
          show ?thesis
          proof (cases s')
          case True note Hs' = this
            with v'_def mext' mul_ext_impl_sound[of _ ps qs] assms(2)
             mul_ext_stri_imp_nstri[of "\<lambda>s t. f (s, t)" ps qs] have Hns': "ns'" by simp
            from m2_valid Htqs Hpps fE_val f_val v_def v'_def Hns Hs Hns' Hs' mext mext'
             show ?thesis by simp
          next
          case False note Hs' = this
            obtain v2 m3 where ex_dom: "ex_dom_efficient_m fE t ps (p # ps) qs m2 = (v2, m3)"
             by (cases "ex_dom_efficient_m fE t ps (p # ps) qs m2") auto
            with IH2' ex_dom_efficient_m_sound[OF _ ps_in_pps rec m2_valid, of t]
             have ex_dom': "ex_dominating (\<lambda>s t. f (s, t)) t ps (p # ps) qs = v2"
             and m3_valid: "model_P model m3 f" by auto
            obtain s2 ns2 where v2_def: "v2 = (s2, ns2)" by (cases v2) auto
            from ex_dominating_sound[OF ps_in_pps] cond1_cond2
             ex_dom' v2_def have "s2 \<Longrightarrow> ns2" unfolding mul_ext_impl_sound by force
            thus ?thesis using m3_valid Htqs Hpps fE_val f_val v_def v'_def v2_def Hns Hs Hs'
             mext mext' ex_dom ex_dom' by simp
          qed
        qed
      next
      case False note Hns = this
        show ?thesis
        proof (cases s)
        case False note Hs = this
          from IH2' ex_dom_efficient_m_sound[OF _ ps_in_pps rec m'_valid, of t]
           show ?thesis using m'_valid Htqs Hpps fE_val f_val v_def Hns Hs by simp
        next
        case True note Hs = this
          obtain ys m2 where filt: "filter_not_stri_efficient_m (fE p) qs m' = (ys, m2)"
           by (cases "filter_not_stri_efficient_m (fE p) qs m'") auto         
          from filt filter_not_stri_efficient_m_sublist[of "fE p" qs m']
           have sub_ys_qs: "sublisteq ys qs" by simp
          from filt filter_not_stri_efficient_m_sound[OF _ m'_valid, of qs "fE p"]
           IH(2)[of p] in_ps_ss in_qs_ts Htqs Hpps
           have m2_valid: "model_P model m2 f" and filt': "ys = [y\<leftarrow> qs. \<not> fst (f (p, y))]" by auto
          obtain v' m3 where mext: "mul_ext_efficient_m fE ps ys m2 = (v', m3)"
           by (cases "mul_ext_efficient_m fE ps ys m2") auto
          from IH(4) Hpps sublisteq_trans list_emb.list_emb_Cons[of "op =" ps ps p] have sub_ps_ss: "sublisteq ps ss" by simp
          from IH(1)[OF _ m2_valid sub_ps_ss sublisteq_trans[OF sub_ys_qs list_emb_Cons2(3)]] mext filt'
           IH(2) in_ps_ss in_qs_ts Htqs Hpps
           have mext': "mul_ext_impl (\<lambda>s t. f (s, t)) ps [y\<leftarrow> qs. \<not> fst (f (p, y))] = v'"
           and m3_valid: "model_P model m3 f"
           by auto
          obtain s' ns' where v'_def: "v' = (s', ns')" by (cases v') auto
          show ?thesis
          proof (cases ns')
          case True note Hns' = this
            from Htqs Hpps fE_val f_val v_def v'_def Hns Hns' Hs filt mext mext' m3_valid
             show ?thesis by simp
          next
          case False note Hns' = this
            from IH2' ex_dom_efficient_m_sound[OF _ ps_in_pps rec m3_valid, of t]
             Htqs Hpps fE_val f_val v_def v'_def Hns Hns' Hs filt mext mext'
             show ?thesis by simp
          qed
        qed
      qed
    qed
  qed
qed

lemma mul_ext_efficient_m_sound:
  assumes "\<And>t s m.  s \<in> set ps \<and> t \<in> set qs \<and> model_P model m f \<Longrightarrow> fst (fE s t m) = f (s, t)
   \<and> model_P model (snd (fE s t m)) f"
  and "model_P model m f"
  shows "model_P model (snd (mul_ext_efficient_m fE ps qs m)) f
   \<and> fst (mul_ext_efficient_m fE ps qs m) = mul_ext_impl (\<lambda>s t. f (s, t)) ps qs"
using assms mul_ext_efficient_m_sound_ind[OF _ _ sublisteq_refl[of ps] sublisteq_refl[of qs]]
by simp

(** The efficient version of RPO is sound *)

lemma m_rpo_valid_sound:
  assumes "m_rpo_valid pr tag m"
  and "memory.lookup model m (p,q) = Some v"
  shows "rpo_unbounded_impl pr tag p q = v"
using assms lookup_valid[of m "\<lambda>p. rpo_unbounded_impl pr tag (fst p) (snd p)"]
unfolding m_rpo_valid_def valid_def by (cases v) auto

lemma store_rpo_sound :
  assumes "m_rpo_valid pr tag m"
  and "rpo_unbounded_impl pr tag s t = v"
  shows "m_rpo_valid pr tag (memory.store model m ((s,t),v))"
using assms store_valid[of m "\<lambda>p. rpo_unbounded_impl pr tag (fst p) (snd p)"]
unfolding m_rpo_valid_def by auto

lemma rpo_efficient_sound:
 assumes "m_rpo_valid pr tag m"
 shows "m_rpo_valid pr tag (snd (rpo_efficient pr tag p q m))
  \<and> fst (rpo_efficient pr tag p q m) = rpo_unbounded_impl pr tag p q"
proof (cases "memory.lookup model m (p, q)")
case (Some v) thus ?thesis using assms m_rpo_valid_sound by simp
next
case None thus ?thesis using assms
  proof (induct p arbitrary: q m)
  case (Var x q m) thus ?case using store_rpo_sound by (cases q) auto
  next
  case (Fun f ss q m) note IH = this
    {
      fix s m assume s_in_ss: "s \<in> set ss" and m_valid: "m_rpo_valid pr tag m"
      have "fst (rpo_efficient pr tag s q m) = rpo_unbounded_impl pr tag s q
       \<and> m_rpo_valid pr tag (snd (rpo_efficient pr tag s q m))"
      proof (cases "memory.lookup model m (s, q)")
      case (Some v) thus ?thesis using m_valid m_rpo_valid_sound by simp
      next
      case None from IH(1)[OF s_in_ss None m_valid] show ?thesis by simp
      qed
    }
    hence rec: "\<And>s m. s \<in> set ss \<Longrightarrow> m_rpo_valid pr tag m \<Longrightarrow>
     fst (rpo_efficient pr tag s q m) = rpo_unbounded_impl pr tag s q
       \<and> m_rpo_valid pr tag (snd (rpo_efficient pr tag s q m))" by simp
    thus ?case using IH(2) IH(3)
    proof (induct q arbitrary: m)
    case (Var y m) note Hq = this
      obtain v m' where any_ns: "any_nstri_efficient_m (\<lambda>s. rpo_efficient pr tag s (Var y))
       ss m = (v, m')" by (cases "any_nstri_efficient_m (\<lambda>s. rpo_efficient pr tag s (Var y)) ss m") auto
      with any_nstri_efficient_m_sound[of ss "\<lambda>f M. model_P model M (\<lambda>p. f (fst p) (snd p))",
       OF _ Hq(3)[unfolded m_rpo_valid_def], of "\<lambda>s. rpo_efficient pr tag s (Var y)" "Var y"] Hq(1)
       have any_ns': "(\<exists>s\<in>set ss. snd (rpo_unbounded_impl pr tag s (Var y))) = v"
       and m'_valid: "m_rpo_valid pr tag m'" unfolding m_rpo_valid_def by auto
      show ?case using Hq(2) any_ns any_ns' m'_valid store_rpo_sound by simp
    next
    case (Fun g ts) note Hq = this
      obtain any m' where any_ns: "any_nstri_efficient_m (\<lambda>s. rpo_efficient pr tag s (Fun g ts))
       ss m = (any, m')"
       by (cases "any_nstri_efficient_m (\<lambda>s. rpo_efficient pr tag s (Fun g ts)) ss m") auto
      with any_nstri_efficient_m_sound[of ss "\<lambda>f M. model_P model M (\<lambda>p. f (fst p) (snd p))",
       OF _ Hq(4)[unfolded m_rpo_valid_def], of "\<lambda>s. rpo_efficient pr tag s (Fun g ts)"] Hq(2)
       have any_ns': "(\<exists>s\<in>set ss. snd (rpo_unbounded_impl pr tag s (Fun g ts))) = any"
        and m'_valid: "m_rpo_valid pr tag m'" unfolding m_rpo_valid_def by auto
      show ?case
      proof (cases any)
      case True note Hany = this
        thus ?thesis using Hq(3) Hq(4) any_ns any_ns' m'_valid store_rpo_sound by simp
      next
      case False note Hany = this
        let ?m = "length ss"
        let ?n = "length ts"
        obtain s ns where Hpr: "fst pr (f, ?m) (g,?n) = (s, ns)" by force
        show ?thesis
        proof (cases ns)
        case False with Hq(3) any_ns any_ns' Hany Hpr m'_valid store_rpo_sound show ?thesis by simp
        next
        case True note Hns = this
          obtain all m2 where all_s: "all_stri_efficient_m (rpo_efficient_m model pr tag
           (Fun f ss)) ts m' = (all, m2)"
           by (cases "all_stri_efficient_m (rpo_efficient_m model pr tag (Fun f ss)) ts m'") auto
          {
            fix t m
            assume t_in_ts: "t \<in> set ts" and m_valid: "m_rpo_valid pr tag m"
            have "fst (rpo_efficient pr tag (Fun f ss) t m) = rpo_unbounded_impl pr tag (Fun f ss) t
             \<and> m_rpo_valid pr tag (snd (rpo_efficient pr tag (Fun f ss) t m))"
            proof (cases "memory.lookup model m (Fun f ss, t)")
            case (Some v) thus ?thesis using m_valid m_rpo_valid_sound by simp
            next
            case None
              {
                fix s m'
                assume s_in_ss: "s \<in> set ss" and m'_valid: "m_rpo_valid pr tag m'"
                have "fst (rpo_efficient pr tag s t m') = rpo_unbounded_impl pr tag s t \<and>
                 m_rpo_valid pr tag (snd (rpo_efficient pr tag s t m'))"
                proof (cases "memory.lookup model m' (s, t)")
                case (Some v) thus ?thesis using m'_valid m_rpo_valid_sound by simp
                next
                case None from IH(1)[OF s_in_ss None m'_valid] show ?thesis by simp
                qed
              }
              with Hq(1)[OF t_in_ts _ None m_valid] IH(1) show ?thesis by simp
            qed  
          }
          with all_stri_efficient_m_sound[of ts "\<lambda>f M. model_P model M (\<lambda>p. f (fst p) (snd p))",
           OF _ m'_valid[unfolded m_rpo_valid_def], of "rpo_efficient pr tag (Fun f ss)"] all_s
           have all_s': "(\<forall>t\<in>set ts. fst (rpo_unbounded_impl pr tag (Fun f ss) t)) = all"
            and m2_valid: "m_rpo_valid pr tag m2" unfolding m_rpo_valid_def by auto
          show ?thesis
          proof (cases s)
          case True
            with Hq(3) any_ns any_ns' all_s all_s' Hany Hns Hpr m2_valid store_rpo_sound
             show ?thesis by simp
          next
          case False note Hs = this
            show ?thesis
            proof (cases all)
            case False note Hall = this
              from Hq(3) any_ns any_ns' all_s all_s' Hall Hany Hns Hpr Hs store_rpo_sound
               m2_valid show ?thesis by (simp add: all_s' Hall)
            next
            case True note Hall = this
              let ?cf = "tag (f, ?m)"
              let ?cg = "tag (g, ?n)"
              show ?thesis
              proof (cases "?cf = ?cg")
              case False
                with Hq(3) any_ns any_ns' all_s all_s' Hall Hany Hns Hs Hpr m2_valid store_rpo_sound
                 show ?thesis by (cases ts, simp_all)
              next
              case True note cfg = this
                {
                  fix t s m
                  assume s_in_ss: "s \<in> set ss" and t_in_ts: "t \<in> set ts"
                   and m_valid: "m_rpo_valid pr tag m"
                  have "fst (rpo_efficient pr tag s t m) = rpo_unbounded_impl pr tag s t
                   \<and> m_rpo_valid pr tag (snd (rpo_efficient pr tag s t m))"
                  proof (cases "memory.lookup model m (s, t)")
                  case (Some v) thus ?thesis using m_valid m_rpo_valid_sound by simp
                  next
                  case None from IH(1)[OF s_in_ss None m_valid] show ?thesis by simp
                  qed
                } note rec = this
                show ?thesis
                proof (cases ?cf)
                case Mul note cf = this
                  obtain v m3 where mext: "mul_ext_efficient_m (rpo_efficient pr tag) ss ts m2
                   = (v, m3)" by (cases "mul_ext_efficient_m (rpo_efficient pr tag) ss ts m2") auto
                  from rpo_unbounded_impl_sound[of "pr" tag] rpo_unbounded_stri_imp_nstri[of "pr" tag]
                   have Himp: "\<forall> x y. fst (rpo_unbounded_impl pr tag x y) \<longrightarrow> snd (rpo_unbounded_impl
                   pr tag x y)" by metis
                  from mul_ext_efficient_m_sound[of ss ts "\<lambda>p. rpo_unbounded_impl pr tag (fst p) (snd p)"
                   "rpo_efficient pr tag" m2] Himp m2_valid[unfolded m_rpo_valid_def] mext rec
                   have mext': "mul_ext_impl (rpo_unbounded_impl pr tag) ss ts = v"
                   and m3_valid: "m_rpo_valid pr tag m3" unfolding m_rpo_valid_def by auto
                  with Hq(3) any_ns any_ns' all_s all_s' Hall Hany Hns Hs Hpr cfg cf mext
                   and store_rpo_sound show ?thesis unfolding mul_ext_impl_sound by simp
                next
                case Lex note cf = this
                  obtain v m3 where lext: "lex_ext_efficient_m (rpo_efficient pr tag) ss ts m2
                   = (v, m3)" by (cases "lex_ext_efficient_m (rpo_efficient pr tag) ss ts m2") auto
                  from lex_ext_efficient_m_sound[OF _ m2_valid[unfolded m_rpo_valid_def],
                   of ss ts "rpo_efficient pr tag"] lext rec
                   have lext': "lex_ext_unbounded (rpo_unbounded_impl pr tag) ss ts = v"
                   and m3_valid: "m_rpo_valid pr tag m3" unfolding m_rpo_valid_def by auto
                  with Hq(3) any_ns any_ns' all_s all_s' Hall Hany Hns Hs Hpr cfg cf lext
                   and store_rpo_sound show ?thesis by simp
                qed
              qed
            qed
          qed
        qed
      qed
    qed
  qed
qed

definition memoized_rpo where
  "memoized_rpo pr tag p q \<equiv> fst (rpo_efficient pr tag p q (memory.empty model ()))"

lemma memoized_rpo_sound:
  "memoized_rpo = rpo_unbounded_impl"
proof-
{
  fix "pr" tag p q
  from rpo_efficient_sound[of "pr" tag "memory.empty model ()"]
   model_satisfactory[unfolded satisfactory_def]
   have "memoized_rpo pr tag p q = rpo_unbounded_impl pr tag p q"
   unfolding m_rpo_valid_def memoized_rpo_def
   by simp
} thus ?thesis by (rule_tac HOL.ext)+
qed

end

(** Now that we have a parametrized version of RPO, we can try the algorithm
    with different kinds of memory implementations *)

(** Using Nagele's first model. *)

definition efficient_rpo_1 where
  [code del]: "efficient_rpo_1 = memoized_rpo.memoized_rpo lm"

interpretation memoized_rpo_1 : memoized_rpo lm valid
  where "memoized_rpo.memoized_rpo lm = efficient_rpo_1"
using satisfactory_lm by unfold_locales (auto simp: efficient_rpo_1_def)

theorem rpo_efficient_1_sound: "rpo_unbounded_impl = efficient_rpo_1"
by (simp add: memoized_rpo_1.memoized_rpo_sound efficient_rpo_1_def)

(** Using Nagele's second model. *)

definition efficient_rpo_2 where
  [code del]: "efficient_rpo_2 = memoized_rpo.memoized_rpo l2m"

interpretation memoized_rpo_2 : memoized_rpo l2m l2m_valid
  where "memoized_rpo.memoized_rpo l2m = efficient_rpo_2"
using satisfactory_l2m by (unfold_locales) (simp_all add: efficient_rpo_2_def)

theorem rpo_efficient_2_sound: "rpo_unbounded_impl = efficient_rpo_2"
by (simp add: memoized_rpo_2.memoized_rpo_sound)

declare rpo_efficient_2_sound[code]

end
